#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>

#include "parse_number.h"

void
print_usage(char * name)
{
  printf(
"NAME\n"
"  %s - extract and save chunks from a file\n"
"\n"
"USAGE\n"
"  %s [options] <source path> <extract path>\n"
"\n"
"OPTIONS\n"
"  -s            : start\n"
"  -e            : end\n"
"  -l            : length\n"
"  -r <filepath> : use length of given reference file\n"
"  -x            : remove extracted chunk from source file\n"
"\n"
"  At least 1 and no more than 2 of -s, -e and -l must be set.\n"
"  -r is considered the same as -l in this context.\n"
"  Negative values are allowed and indicate offsets from the end of the file.\n"
"  If only the length is given then it is interpretted as a length from the\n"
"  of the file if positive or a length from the end of the file if negative.\n"
"\n"
"  -x will remove the extracted chunk from the source file and shift the\n"
"  contents at the end to fill in the excised chunk. This is intended for\n"
"  managing one-time pads.\n"
"\n"
"  Offsets and lengths may include a prefix (k,M,G,T,P,E,Z,Y) and may be\n"
"  followed by \"B\" or \"iB\". If followed by \"B\" the prefix is interpretted\n"
"  as an SI prefix (e.g. \"1kB\" = 1000 bytes), otherwise it is interpretted as\n"
"  a binary prefix (e.g. \"1k\" = \"1kiB\" = 1024 bytes).\n"
"\n"
"EXAMPLES\n"
"  Extract a chunk from a one-time pad named \"pad\" to a file named \"page\"\n"
"  and use it to encrypt a file named \"foo\" with otsplit. The extracted page\n"
"  is removed from the pad.\n"
"\n"
"    %s name -xr foo pad page\n"
"    otsplit -s -p page foo.encrypted < foo\n"
"\n"
"  If the pad has been shared with someone securely (e.g. transferred in\n"
"  person) then the encrypted file can be sent to that person for decryption via\n"
"  an insecure channel. As long as the pad is never comprimised, the data should\n"
"  be safe. (Note however that the filesize will be visible and may be used to\n"
"  inferred the contents of the file. In this is a risk then the file size\n"
"  be modified to disguise it, e.g. by padding it with zeros, compressing it, or\n"
"  breaking it into chunks and sent with dummy chunks.)\n"
"\n"
"  To decrypt the file on the other end, extract a page of the same size used to\n"
"  encrypt the data and then decrypt it with otsplit.\n"
"\n"
"    %s name -xr foo.encrypted pad page\n"
"    otsplit -j -p page foo.encrypted > foo\n"
"\n"
"  Both pads are now in the same state and can be used for further encryption.\n"
"  Extra caution may be taken with the pad by omitting the -x option to first\n"
"  extract a page without removing it from the pad. The command can then be\n"
"  rerun with -x after confirmation of successful decryption.\n"
"\n"
"DISCLAIMER\n"
"  This program is distributed in the hope that it will be useful,\n"
"  but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
"\n"
"  If you require reliable cryptography, check the source code yourself to\n"
"  ensure that it satisfies your requirements.\n"
"\n"
"AUTHOR\n"
"  Xyne, 2013\n",
  name,
  name,
  name,
  name
  );
}


int
main(int argc, char * * argv)
{
  int i, excise;
  long long signed int start, end, length, ret;
  size_t source_file_size;
  int start_set, end_set, length_set;
  char * path_source, * path_extract, * path_reference;
  int fd_source, fd_extract;
  char * map_source, * map_extract;
  struct stat st;

  excise = 0;
  start_set = 0;
  end_set = 0;
  length_set = 0;
  start = 0;
  end = 0;
  length = 0;
  path_reference = NULL;


  while ((i = getopt(argc, argv, "hs:e:l:r:x")) != -1)
  {
    switch (i)
    {
      case 'h':
        print_usage(argv[0]);
        return EXIT_SUCCESS;
      case 's':
        start = parse_number(optarg);
        start_set = 1;
        break;
      case 'e':
        end = parse_number(optarg);
        end_set = 1;
        break;
      case 'l':
        length = parse_number(optarg);
        length_set = 1;
        break;
      case 'r':
        path_reference = optarg;
        break;
      case 'x':
        excise = 1;
        break;
      case '?':
        print_usage(argv[0]);
        return EXIT_FAILURE;
    }
  }

  if (
    argc - optind != 2 ||
    ! (start_set || end_set || length_set || path_reference != NULL)
  )
  {
    print_usage(argv[0]);
    return EXIT_FAILURE;
  }

  if (path_reference != NULL)
  {
    if (stat(path_reference, &st))
    {
      perror("error: failed to stat reference file");
      return EXIT_FAILURE;
    }
   length = st.st_size;
   length_set = 1;
  }

  path_source = argv[optind];
  path_extract = argv[optind+1];

  if (stat(path_source, &st))
  {
    perror("error: failed to stat source file");
    return EXIT_FAILURE;
  }
  source_file_size = st.st_size;

  /*
    Gotta catch'em all!
  */
  if (start_set && end_set && length_set)
  {
    fprintf(stderr, "error: start, end and length may not be set simultaneously\n");
    return EXIT_FAILURE;
  }

  if (start_set)
  {
    if (start < 0)
    {
      start = source_file_size + start;
      if (start < 0)
      {
        fprintf(stderr, "error: source file size + start < 0\n");
        return EXIT_FAILURE;
      }
    }
    else
    {
      if (start > source_file_size)
      {
        fprintf(stderr, "error: start > source file size\n");
        return EXIT_FAILURE;
      }
    }
  }

  if (end_set)
  {
    if (end < 0)
    {
      end = source_file_size + end;
      if (end < 0)
      {
        fprintf(stderr, "error: source file size + end < 0\n");
        return EXIT_FAILURE;
      }
    }
    else
    {
      if (end > source_file_size)
      {
        fprintf(stderr, "error: end > source file size\n");
        return EXIT_FAILURE;
      }
    }
  }

  if (start_set && end_set)
  {
    length = end - start;
    if (length < 1)
    {
      fprintf(stderr, "error: end - start < 1\n");
      return EXIT_FAILURE;
    }
  }

  else if (start_set)
  {
    if (length_set)
    {
      end = start + length;
      if (end <= start)
      {
        fprintf(stderr, "error: start + length <= start\n");
        return EXIT_FAILURE;
      }
      else if (end > source_file_size)
      {
        fprintf(stderr, "error: end > source file size\n");
        return EXIT_FAILURE;
      }
    }
    else if (start == 0)
    {
      fprintf(stderr, "error: extract is entire source file\n");
      return EXIT_FAILURE;
    }
  }

  else if (end_set)
  {
    if (length_set)
    {
      start = end - length;
      if (start < 0)
      {
        fprintf(stderr, "error: start < 0\n");
        return EXIT_FAILURE;
      }
      else if (start >= end)
      {
        fprintf(stderr, "error: start >= end\n");
        return EXIT_FAILURE;
      }
    }
    else if (end == source_file_size)
    {
      fprintf(stderr, "error: extract is entire source file\n");
      return EXIT_FAILURE;
    }
  }

  else if (length == 0)
  {
    fprintf(stderr, "error: length == 0\n");
    return EXIT_FAILURE;
  }
  else if (abs(length) >= source_file_size)
  {
    fprintf(stderr, "error: abs(length) >= source_file_size\n");
    return EXIT_FAILURE;
  }
  else if (length > 0)
  {
    start = 0;
    end = length;
  }
  else
  {
    end = source_file_size;
    start = end + length;
    length = -length;
  }





  fd_source = open(path_source, O_RDWR, (mode_t) 0600);
  if (fd_source == -1)
  {
    perror("error: failed to open source file");
    return EXIT_FAILURE;
  }

//   ret = lseek(fd_source, 0L, SEEK_END);
//   if (ret == -1)
//   {
//     close(fd_source);
//     perror("error: failed to seek to end of source file");
//     return EXIT_FAILURE;
//   }


  fd_extract = open(path_extract, O_RDWR | O_CREAT | O_TRUNC | O_APPEND, (mode_t) 0600);
  if (fd_extract == -1)
  {
    perror("error: failed to open source file");
    return EXIT_FAILURE;
  }

  ret = lseek(fd_extract, length, SEEK_SET);
  if (ret == -1)
  {
    close(fd_source);
    close(fd_extract);
    perror("error: failed to seek to end of extract file");
    return EXIT_FAILURE;
  }

  ret = write(fd_extract, "", 1);
  if (ret == -1)
  {
    close(fd_source);
    close(fd_extract);
    perror("error: failed to write to end of extract file");
    return EXIT_FAILURE;
  }



  map_source = (char *) mmap(0, source_file_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd_source, 0);
  if (map_source == MAP_FAILED)
  {
    close(fd_source);
    close(fd_extract);
    perror("error: failed to mmap source file");
    return EXIT_FAILURE;
  }

  map_extract = (char *) mmap(0, length, PROT_WRITE, MAP_SHARED, fd_extract, 0);
  if (map_extract == MAP_FAILED)
  {
    perror("error: failed to mmap extract file");
    if (munmap(map_source, source_file_size))
    {
      perror("error: failed to munmap source file");
    }
    close(fd_source);
    close(fd_extract);
    return EXIT_FAILURE;
  }

  errno = 0;
  memcpy(map_extract, map_source + start, length);
  if (errno)
  {
    perror("error: failed to extract source data");
  }
  else if (ftruncate(fd_extract, length))
  {
    perror("error: failed to truncate extracted file");
  }
  else if (excise)
  {
    memmove(map_source + start, map_source + end, source_file_size - end);
    if (errno)
    {
      perror("error: failed to shift data in source file after extraction");
    }
    else if (ftruncate(fd_source, source_file_size - length))
    {
      perror("error: failed to truncate source file after extraction");
    }
  }

  if (munmap(map_source, source_file_size))
  {
    perror("error: failed to munmap source file");
  }
  if (munmap(map_extract, length))
  {
    perror("error: failed to munmap extract file");
  }
  close(fd_source);
  close(fd_extract);
  return EXIT_SUCCESS;
}
